/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS    Arch Lab. TOKYO TECH *
 **********************************************************************/
#include "define.h"
#include <ncurses.h>

#include <termios.h>
#include <unistd.h>
#include <errno.h>

/**********************************************************************/
Display::Display()
{
    board = NULL;
    window_width = window_height = 0;
    lcd_width = lcd_height = 0;
    cursorx = cursory = 0;
    lcdindex = 0;
    gui_mode = 0;
}

/**********************************************************************/
void Display::lcd_ttyopen()
{

    initscr();

    if(gui_mode){
        
        return;
    }

    start_color();
    if (COLORS >= 8) {
        init_pair(1, COLOR_YELLOW, COLOR_BLACK);
        init_pair(2, COLOR_MAGENTA, COLOR_BLACK);
        init_pair(3, COLOR_RED, COLOR_BLACK);
        init_pair(4, COLOR_CYAN, COLOR_BLACK);
        init_pair(5, COLOR_GREEN, COLOR_BLACK);
        init_pair(6, COLOR_BLUE, COLOR_BLACK);
        init_pair(7, COLOR_BLACK, COLOR_BLACK);
    }
    
    clear();

    cursorx = cursory = 0;
    
    for (int i = 0; i < lcd_height; i++) {
        move(i, lcd_width);
        printw("#");
    }
    move(lcd_height, 0);
    for (int i = 0; i < lcd_width + 1; i++)
        printw("#");
    
    refresh();
}

/**********************************************************************/
void Display::lcd_ttyclose()
{
    endwin();
}

/**********************************************************************/
void Display::lcd_setcolor(int color)
{
    attrset(COLOR_PAIR(((color & 0x80) >> 5) ^
                       ((color & 0x10) >> 3) ^
                       ((color & 0x02) >> 1) ^ 0x7));
}

/**********************************************************************/
void Display::lcd_cls()
{
    int x, y;
    for (y = 0; y < lcd_height; y++) {
        move(y, 0);
        for (x = 0; x < lcd_width; x++)
            printw(" ");
    }

    refresh();
}

/**********************************************************************/
int Display::lcd_puts(char *str)
{
    char *p = str;

    move(cursory, cursorx);
    while(*p != 0){
        if( 0x20 < *p && *p < 0x7f)
            printw("%c", *p);
        else
            printw(" ");

        if (++cursorx >= lcd_width) {
            cursorx = 0;
            cursory = (cursory + 1) % lcd_height;
            move(cursory, cursorx);
        }
        p++;
    }
    
    refresh();
    return p-str;
}

/**********************************************************************/
int Display::lcd_debugprintf(char *fmt, ...)
{
    char buf[1024], *p;
    int ret;
    va_list arg;
    va_start(arg, fmt);

    ret = vsnprintf(buf, 1024, fmt, arg);
    va_end(arg);

    move(lcd_height+1, 0);
    p = buf;
    while(*p != 0){
        if( 0x20 < *p && *p < 0x7f)
            printw("%c", *p);
        else
            printw(" ");

        if (++cursorx >= lcd_width) {
            move(lcd_height + 2, 0);
        }
        p++;
    }
    return ret;
}

/**********************************************************************/
void Display::lcd_locate(int x, int y)
{
    cursorx = x % lcd_width;
    cursory = y % lcd_height;
}

/**********************************************************************/
void Display::lcd_nextline()
{
    cursorx = 0;
    cursory = (cursory + 1) % lcd_height;
}


/**********************************************************************/
int Display::lcd_printf(char *fmt, ...)
{
    char buf[1024];
    int ret;
    va_list arg;
    va_start(arg, fmt);

    move(cursory, cursorx);
    ret = vsnprintf(buf, 1024, fmt, arg);
    va_end(arg);

    lcd_puts(buf);
    return ret;
}

/**********************************************************************/
void Lcdv1::init(Board *board)
{
    int width, height;

    this->board = board;

    if(board->lcd_rotate){
        width  = 320;
        height = 240;
    }else{
        width  = 240;
        height = 320;
    }

    this->window_width = width;
    this->window_height = height;
    this->lcd_width = window_width / 8;
    this->lcd_height = window_height / 16;

    lcd_ttyopen();
    lcdindex = 0;
}

/**********************************************************************/
void Lcdv1::fini()
{
    lcd_ttyclose();
}

/**********************************************************************/
void Lcdv1::send(char data)
{
    if(lcdindex < LCD_BUF_SIZE - 1){
        lcdbuf[lcdindex++] = data;
    }
}

/**********************************************************************/
void Lcdv1::step()
{    
    if( lcdbuf[lcdindex - 1] == '\r' ){
        lcdbuf[lcdindex - 1] = 0;
        
        parse(lcdbuf, lcdindex - 1);
        lcdindex = 0;
    }
}

/**********************************************************************/
int Lcdv1::parse(char *cmd, int length)
{
    int x, y;

    if (strncmp(cmd, "CS", 2) == 0) {
        lcd_setcolor(strtol(cmd + 2, NULL, 16));
    } else if (strncmp(cmd, "ER", 2) == 0) {
        lcd_cls();
        lcd_locate(0,0);
    } else if (strncmp(cmd, "HP", 2) == 0) {
        char *endptr;
        x = strtol(cmd + 2, &endptr, 10);
        y = strtol(endptr + 1, NULL, 10);
        lcd_locate(x, y);
    } else if (strncmp(cmd, "HW", 2) == 0) {
        cmd[length - 1] = 0;
        lcd_puts(cmd + 3);
    } else if (strncmp(cmd, "HR", 2) == 0) {
        lcd_nextline();
    }

    return 0;
}

/**********************************************************************/
void Lcdv2::init(Board *board)
{
    int width, height;
    int pid;
    int ret;
    char buf[32];

    this->board = board;
    if(board->lcd_type == 1){
        if(board->lcd_rotate){
            width  = 320;
            height = 240;
        }else{
            width  = 240;
            height = 320;
        }

    }else{
        if(board->lcd_rotate){
            width  = 480;
            height = 272;
        }else{
            width  = 272;
            height = 480;
        }
    }
    this->window_width = width;
    this->window_height = height;
    this->lcd_width = window_width / 8;
    this->lcd_height = window_height / 16;

    gui_mode = board->lcd_mode;
    if(gui_mode){ 
        ret = pipe(gui_pipe);
        if(ret == -1){
            fprintf(stderr, "## Can't open LCD pipe.\n");
            exit(1);
        }
        
        pid = fork();
        if(pid == 0){
            close(gui_pipe[1]);
            close(0);
            dup2(gui_pipe[0], 0);
            close(gui_pipe[0]);
            
            snprintf(buf, 32, "-w%d -h%d", width, height);
            ret = execlp(board->lcd_path, board->lcd_path, buf , NULL);
            
            if(ret == -1){
                fprintf(stderr,"## Can't open LCD (%s).\n",board->lcd_path);
                exit(1);
            }
            exit(0);
        }else{
            usleep(1000000);
        }
    }

    lcd_ttyopen();
    lcdindex = 0;
}

/**********************************************************************/
void Lcdv2::fini()
{
    if(gui_mode)
        close(gui_pipe[1]);
    
    lcd_ttyclose();
}

/**********************************************************************/
void Lcdv2::send(char data)
{
    if(lcdindex < LCD_BUF_SIZE - 1){
        lcdbuf[lcdindex++] = data;
    }
}

/**********************************************************************/
void Lcdv2::step()
{
    static int count = 0;

    if(gui_mode){
        count++;
        if(count%100000 == 0 || lcdindex > LCD_BUF_SIZE - 100){
            if(lcdindex > 0){
                sendcmd(lcdbuf, lcdindex);
                lcdindex = 0;
            }else{
                sendcmd("", 1);
            }
                count = 0;
        }
        return;
    }


    if(parse(lcdbuf, lcdindex))
        lcdindex = 0;
}

/**********************************************************************/
int Lcdv2::parse(char *cmd, int length)
{
    switch(cmd[0]){
    case 'C':
        if(length < 2) return 0;
        if(chartoint(cmd[1]) == 0)
            lcd_cls();
        break;
                
    case 'W':
        if(length < 2) return 0;
        // do nothing
        break;
            
    case 'P':
        if(length < 2) return 0;
        // do nothing
        break;
            
    case 'R':
        if(length < 2) return 0;
        // do nothing
        break;
            
    case 'T':
        if(length < 2) return 0;
        // do nothing
        break;
            
    case 'F':
        if(length < 4) return 0;
        {
            int r,g,b;
            r = chartoint(cmd[1]);
            g = chartoint(cmd[2]);
            b = chartoint(cmd[3]);
            lcd_setcolor(((r&0x7)<<5) | ((g&0x7)<<2) | (b&0x3));
        }
        break;
            
    case 'B':
        if(length < 4) return 0;
        // do nothing
        break;
            
    case 'H':
        if(length < 2) return 0;
        lcd_printf("%c", cmd[1]);
        break;
                
    case 'A':
        if(length < 2) return 0;
        // do nothing
        break;

    case 'Z':
        if(length < 5) return 0;
        // do nothing
        break;

    case 'X':
        if(length < 3) return 0;
        {
            int n;
            n = chartoint(cmd[1]);
            n = chartoint(cmd[2]) + n * 10;
            lcd_locatex(n);
        }
        break;

    case 'Y':
        if(length < 3) return 0;
        {
            int n;
            n = chartoint(cmd[1]);
            n = chartoint(cmd[2]) + n * 10;
            lcd_locatey(n/2);
        }
        break;

    case 'M':
        if(length < 2) return 0;
        lcd_locatex(cursorx + chartoint(cmd[1]));
        break;

    case 'N':
        if(length < 2) return 0;
        lcd_locatey(cursory + chartoint(cmd[1])/2);
        break;
                
    case 'D':
        if(length < 1) return 0;
        // do nothing
        break;

    case 'I':
        if(length < 4) return 0;
        // do nothing
        break;

    case 'J':
        if(length < 4) return 0;
        // do nothing
        break;

    case 'K':
        if(length < 2) return 0;
        // do nothing
        break;

    case 'L':
        if(length < 2) return 0;
        // do nothing
        break;

    case 'V':
        if(length < 3) return 0;
        // do nothing
        break;

    case 'S':
        if(length < 5) return 0;
        // do nothing
        break;

    case 0xf5:
        if(length < 2) return 0;
        lcd_locatex(cursorx - chartoint(cmd[1]));
        break;

    case 0xf6:
        if(length < 2) return 0;
        lcd_locatey(cursory - chartoint(cmd[1]));
        break;
                
    default: ;
        // do nothing
    }

    return 1;

}

/**********************************************************************/
void Lcdv2::sendcmd(char *cmd, int length)
{
    write(gui_pipe[1], cmd ,length);
}

/**********************************************************************/
int Lcdv2::chartoint(char c)
{
    if('0' <= c && c <= '9')
        return c - '0';
    if('a' <= c && c <= 'f')
        return c - 'a';
    if('A' <= c && c <= 'F')
        return c - 'A';
    return -1;
}

/**********************************************************************/
void Lcdv2::lcd_locatex(int x)
{
    if(0 <= x && x < lcd_width)
        cursorx = x;
}

/**********************************************************************/
void Lcdv2::lcd_locatey(int y)
{
    if(0 <= y && y < lcd_height)
        cursory = y;
}


/**********************************************************************/
