/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS    Arch Lab. TOKYO TECH *
 **********************************************************************/
#include "define.h"

inline uint032_t exts32(uint032_t x, int y);
inline const char *setinst(uint);
#define READ_FCSR(pos) ((fcsr << FCSR_ ## pos) & 0x1)
#define WRITE_FCSR(pos, val) fcsr = ((fcsr & ~(1 << FCSR_ ## pos)) |\
                                     ((val) << FCSR_ ## pos))

/**********************************************************************/
enum {
    FCSR_RM = 0,
    FCSR_FLG_I,
    FCSR_FLG_U,
    FCSR_FLG_O,
    FCSR_FLG_Z,
    FCSR_FLG_V,
    FCSR_ENB_I,
    FCSR_ENB_U,
    FCSR_ENB_O,
    FCSR_ENB_Z,
    FCSR_ENB_V,
    FCSR_COZ_I,
    FCSR_COZ_U,
    FCSR_COZ_O,
    FCSR_COZ_Z,
    FCSR_COZ_V,
    FCSR_E = 17,
    /* Three Reserved Bits */
    FCSR_FN = 21,
    FCSR_FO,
    FCSR_FCC0,
    FCSR_FS,
    FCSR_FCC1,
    FCSR_FCC2,
    FCSR_FCC3,
    FCSR_FCC4,
    FCSR_FCC5,
    FCSR_FCC6,
    FCSR_FCC7 = 31,
    FCSR_FEXR_MASK = 0x0003f07f,
    FCSR_FENR_MASK = 0x00000f83,
    FCSR_FCC_MASK  = 0xfe000000,
    FENR_FS = 2,
    FCCR_FCC1 = 1,
};

/**********************************************************************/
MipsFpu::MipsFpu(Board *board)
{
    this->board = board;
    for (int i = 0; i < NFREG; i++)
        r.w[i] = 0;
    fcsr = 0;
}

/**********************************************************************/
int MipsFpu::execute(MipsInst *inst, MipsCp0 *cp0, uint032_t *rrs, 
                     uint032_t *rrt, uint032_t *rrd, uint032_t *npc,
                     uint032_t *vaddr, uint064_t *paddr, uint064_t *r64,
                     int *cond)
{
    uint fcc[8] = {1 << FCSR_FCC0, 1 << FCSR_FCC1, 
                   1 << FCSR_FCC2, 1 << FCSR_FCC3, 
                   1 << FCSR_FCC4, 1 << FCSR_FCC5, 
                   1 << FCSR_FCC6, 1 << FCSR_FCC7};
    uint c_bit = 0;
    int ret;
    
    switch (inst->fop) {
    case MOVF_____:
    case MOVT_____:
        *cond = ((fcsr & fcc[inst->rt >> 2]) == 0) ^ (inst->rt & 0x1);
        *rrd = *rrs;
        break;
    case MFC1_____:
        *rrt = r.w[inst->rd];
        break;
    case CFC1_____:
        if (inst->rd == FPU_FIR_) {
            *rrt = FIR_VALUE;
        } else if (inst->rd == FPU_FCSR) {
            *rrt = fcsr;
        } else if (inst->rd == FPU_FCCR) {
            *rrt = (fcsr >> (FCSR_FCC1 - FCCR_FCC1)) | READ_FCSR(FCC0);
        } else if (inst->rd == FPU_FEXR) {
            *rrt = fcsr & FCSR_FEXR_MASK;
        } else if (inst->rd == FPU_FENR) {
            *rrt = (fcsr & FCSR_FENR_MASK) | (READ_FCSR(FS) << FENR_FS);
        } else {
            *rrt = 0;
        }
        break;
    case MTC1_____:
        r.w[inst->rd] = *rrt;
        break;
    case CTC1_____:
        if (inst->rd == FPU_FCSR) {
            fcsr = *rrt;
        } else if (inst->rd == FPU_FCCR) {
            fcsr = ((fcsr & ~FCSR_FCC_MASK) | 
                    ((*rrt << (FCSR_FCC1 - FCCR_FCC1)) & FCSR_FCC_MASK));
            WRITE_FCSR(FCC0, *rrt & 0x1);
        } else if (inst->rd == FPU_FEXR) {
            fcsr = ((fcsr & ~FCSR_FEXR_MASK) | (*rrt & FCSR_FEXR_MASK));
        } else if (inst->rd == FPU_FENR) {
            fcsr = ((fcsr & ~FCSR_FENR_MASK) | (*rrt & FCSR_FENR_MASK));
            WRITE_FCSR(FS, (*rrt >> FENR_FS) & 0x1);
        }
        break;
    case BC1F_____:
    case BC1FL____:
        *cond = ((fcsr & fcc[inst->rt >> 2]) == 0);
        *npc = inst->pc + (exts32(inst->imm, 16) << 2) + 4;
        break;
    case BC1T_____:
    case BC1TL____:
        *cond = ((fcsr & fcc[inst->rt >> 2]) != 0);
        *npc = inst->pc + (exts32(inst->imm, 16) << 2) + 4;
        break;
    case ADD_S____:
        r.s[inst->shamt] = r.s[inst->rd] + r.s[inst->rt];
        break;
    case ADD_D____:
        r.d[inst->shamt / 2] = r.d[inst->rd / 2] + r.d[inst->rt / 2];
        break;
    case SUB_S____:
        r.s[inst->shamt] = r.s[inst->rd] - r.s[inst->rt];
        break;
    case SUB_D____:
        r.d[inst->shamt / 2] = r.d[inst->rd / 2] - r.d[inst->rt / 2];
        break;
    case MUL_S____:
        r.s[inst->shamt] = r.s[inst->rd] * r.s[inst->rt];
        break;
    case MUL_D____:
        r.d[inst->shamt / 2] = r.d[inst->rd / 2] * r.d[inst->rt / 2];
        break;
    case DIV_S____:
        r.s[inst->shamt] = r.s[inst->rd] / r.s[inst->rt];
        break;
    case DIV_D____:
        r.d[inst->shamt / 2] = r.d[inst->rd / 2] / r.d[inst->rt / 2];
        break;
    case SQRT_S___:
        r.s[inst->shamt] = sqrtf(r.s[inst->rd]);
        break;
    case SQRT_D___:
        r.d[inst->shamt / 2] = sqrt(r.d[inst->rd / 2]);
        break;
    case ABS_S____:
        r.s[inst->shamt] = fabsf(r.s[inst->rd]);
        break;
    case ABS_D____:
        r.d[inst->shamt / 2] = fabs(r.d[inst->rd / 2]);
        break;
    case MOV_S____:
        r.s[inst->shamt] = r.s[inst->rd];
        break;
    case MOV_D____:
        r.d[inst->shamt / 2] = r.d[inst->rd / 2];
        break;
    case NEG_S____:
        r.s[inst->shamt] = -r.s[inst->rd];
        break;
    case NEG_D____:
        r.d[inst->shamt / 2] = -r.d[inst->rd / 2];
        break;
    case ROUND_W_S:
        r.w[inst->shamt] = (int032_t) roundf(r.s[inst->rd]);
        break;
    case ROUND_W_D:
        r.w[inst->shamt] = (int032_t) round(r.d[inst->rd / 2]);
        break;
    case TRUNC_W_S:
        r.w[inst->shamt] = (int032_t) truncf(r.s[inst->rd]);
        break;
    case TRUNC_W_D:
        r.w[inst->shamt] = (int032_t) trunc(r.d[inst->rd / 2]);
        break;
    case CEIL_W_S_:
        r.w[inst->shamt] = (int032_t) ceilf(r.s[inst->rd]);
        break;
    case CEIL_W_D_:
        r.w[inst->shamt] = (int032_t) ceil(r.d[inst->rd / 2]);
        break;
    case FLOOR_W_S:
        r.w[inst->shamt] = (int032_t) floorf(r.s[inst->rd]);
        break;
    case FLOOR_W_D:
        r.w[inst->shamt] = (int032_t) floor(r.d[inst->rd / 2]);
        break;
    case MOVF_S___:
        if ((fcsr & fcc[inst->rt >> 2]) == 0)
            r.s[inst->shamt] = r.s[inst->rd];
        break;
    case MOVF_D___:
        if ((fcsr & fcc[inst->rt >> 2]) == 0)
            r.d[inst->shamt / 2] = r.d[inst->rd / 2];
        break;
    case MOVT_S___:
        if ((fcsr & fcc[inst->rt >> 2]) != 0)
            r.s[inst->shamt] = r.s[inst->rd];
        break;
    case MOVT_D___:
        if ((fcsr & fcc[inst->rt >> 2]) != 0)
            r.d[inst->shamt / 2] = r.d[inst->rd / 2];
        break;
    case MOVZ_S___:
        if (!*rrt)
            r.s[inst->shamt] = r.s[inst->rd];
        break;
    case MOVZ_D___:
        if (!*rrt)
            r.d[inst->shamt / 2] = r.d[inst->rd / 2];
        break;
    case MOVN_S___:
        if (*rrt)
            r.s[inst->shamt] = r.s[inst->rd];
        break;
    case MOVN_D___:
        if (*rrt)
            r.d[inst->shamt / 2] = r.d[inst->rd / 2];
        break;
    case RECIP_S__:
        r.s[inst->shamt] = (float) 1.0 / r.s[inst->rd];
        break;
    case RECIP_D__:
        r.d[inst->shamt / 2] = (double) 1.0 / r.d[inst->rd / 2];
        break;
    case RSQRT_S__:
        r.s[inst->shamt] = sqrtf((float) 1.0 / r.s[inst->rd]);
        break;
    case RSQRT_D__:
        r.d[inst->shamt / 2] = sqrt((double) 1.0 / r.d[inst->rd / 2]);
        break;
    case CVT_S_D__:
        r.s[inst->shamt] = (float) r.d[inst->rd / 2];
        break;
    case CVT_S_W__:
        r.s[inst->shamt] = (float) (int032_t) r.w[inst->rd];
        break;
    case CVT_D_S__:
        r.d[inst->shamt / 2] = (double) r.s[inst->rd];
        break;
    case CVT_D_W__:
        r.d[inst->shamt / 2] = (double) (int032_t) r.w[inst->rd];
        break;
    case CVT_W_S__:
        r.w[inst->shamt] = (int032_t) r.s[inst->rd];
        break;
    case CVT_W_D__:
        r.w[inst->shamt] = (int032_t) r.d[inst->rd / 2];
        break;
    case C_F_S____:
    case C_SF_S___:
    case C_F_D____:
    case C_SF_D___:
        c_bit = 0;
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_UN_S___:
        c_bit = isnan(r.s[inst->rd]) || isnan(r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_UN_D___:
        c_bit = isnan(r.d[inst->rd / 2]) || isnan(r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_EQ_S___:
    case C_SEQ_S__:
        c_bit = (r.s[inst->rd] == r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_EQ_D___:
    case C_SEQ_D__:
        c_bit = (r.d[inst->rd / 2] == r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_UEQ_S__:
    case C_NGL_S__:
        c_bit = isnan(r.s[inst->rd]) || isnan(r.s[inst->rt]);
        c_bit = c_bit || (r.s[inst->rd] == r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_UEQ_D__:
    case C_NGL_D__:
        c_bit = isnan(r.d[inst->rd / 2]) || isnan(r.d[inst->rt / 2]);
        c_bit = c_bit || (r.d[inst->rd / 2] == r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_OLT_S__:
    case C_LT_S___:
        c_bit = (r.s[inst->rd] < r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_OLT_D__:
    case C_LT_D___:
        c_bit = (r.d[inst->rd / 2] < r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_ULT_S__:
    case C_NGE_S__:
        c_bit = isnan(r.s[inst->rd]) || isnan(r.s[inst->rt]);
        c_bit = c_bit || (r.s[inst->rd] < r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_ULT_D__:
    case C_NGE_D__:
        c_bit = isnan(r.d[inst->rd / 2]) || isnan(r.d[inst->rt / 2]);
        c_bit = c_bit || (r.d[inst->rd / 2] < r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_OLE_S__:
    case C_LE_S___:
        c_bit = (r.s[inst->rd] == r.s[inst->rt]);
        c_bit = c_bit || (r.s[inst->rd] < r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_OLE_D__:
    case C_LE_D___:
        c_bit = (r.d[inst->rd / 2] == r.d[inst->rt / 2]);
        c_bit = c_bit || (r.d[inst->rd / 2] < r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_ULE_S__:
    case C_NGT_S__:
        c_bit = isnan(r.s[inst->rd]) || isnan(r.s[inst->rt]);
        c_bit = c_bit || (r.s[inst->rd] == r.s[inst->rt]);
        c_bit = c_bit || (r.s[inst->rd] < r.s[inst->rt]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case C_ULE_D__:
    case C_NGT_D__:
        c_bit = isnan(r.d[inst->rd / 2]) || isnan(r.d[inst->rt / 2]);
        c_bit = c_bit || (r.d[inst->rd / 2] == r.d[inst->rt / 2]);
        c_bit = c_bit || (r.d[inst->rd / 2] < r.d[inst->rt / 2]);
        fcsr = ((fcsr & ~fcc[inst->shamt >> 2]) | 
                (c_bit * fcc[inst->shamt >> 2]));
        break;
    case LWXC1____:
        *vaddr = *rrs + *rrt;
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x3)
            return EXC_ADEL___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 0)) != 0)
            return ret;
        break;
    case LDXC1____:
        *vaddr = *rrs + *rrt;
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x7)
            return EXC_ADEL___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 0)) != 0)
            return ret;
        break;
    case SWXC1____:
        *vaddr = *rrs + *rrt;
        *rrt = r.w[inst->rd];
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x3)
            return EXC_ADES___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 1)) != 0)
            return ret;
        break;
    case SDXC1____:
        *vaddr = *rrs + *rrt;
        *r64 = r.l[inst->rd / 2];
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x7)
            return EXC_ADES___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 1)) != 0)
            return ret;
        break;
    case MADD_S___:
        r.s[inst->shamt] = (r.s[inst->rs] + 
                            r.s[inst->rd] * r.s[inst->rt]);
        break;
    case MADD_D___:
        r.d[inst->shamt / 2] = (r.d[inst->rs / 2] +
                                r.d[inst->rd / 2] * r.d[inst->rt / 2]);
        break;
    case MSUB_S___:
        r.s[inst->shamt] = (- r.s[inst->rs] + 
                            r.s[inst->rd] * r.s[inst->rt]);
        break;
    case MSUB_D___:
        r.d[inst->shamt / 2] = (- r.d[inst->rs / 2] +
                                r.d[inst->rd / 2] * r.d[inst->rt / 2]);
        break;
    case NMADD_S__:
        r.s[inst->shamt] = (- r.s[inst->rs] - 
                            r.s[inst->rd] * r.s[inst->rt]);
        break;
    case NMADD_D__:
        r.d[inst->shamt / 2] = (- r.d[inst->rs / 2] -
                                r.d[inst->rd / 2] * r.d[inst->rt / 2]);
        break;
    case NMSUB_S__:
        r.s[inst->shamt] = (r.s[inst->rs] - 
                            r.s[inst->rd] * r.s[inst->rt]);
        break;
    case NMSUB_D__:
        r.d[inst->shamt / 2] = (r.d[inst->rs / 2] -
                                r.d[inst->rd / 2] * r.d[inst->rt / 2]);
        break;
    case L_S______:
        *vaddr = *rrs + exts32(inst->imm, 16);
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x3)
            return EXC_ADEL___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 0)) != 0)
            return ret;
        break;
    case L_D______:
        *vaddr = *rrs + exts32(inst->imm, 16);
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x7)
            return EXC_ADEL___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 0)) != 0)
            return ret;
        break;
    case S_S______:
        *vaddr = *rrs + exts32(inst->imm, 16);
        *rrt = r.w[inst->rt];
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x3)
            return EXC_ADES___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 1)) != 0)
            return ret;
        break;
    case S_D______:
        *vaddr = *rrs + exts32(inst->imm, 16);
        *r64 = r.l[inst->rt / 2];
        if (!cp0)
            *paddr = (uint064_t) *vaddr;
        else if (*vaddr & 0x7)
            return EXC_ADES___;
        else if ((ret = cp0->getphaddr(*vaddr, paddr, 1)) != 0)
            return ret;
        break;
    default:
        printf("## Undefined Opcode ! %08x: %08x\n",
               inst->pc, inst->ir);
        return -1;
        break;
    }
    return 0;
}

/**********************************************************************/
void MipsFpu::writeback(MipsInst *inst, uint032_t rrt, uint064_t r64)
{
    if (inst->fop == LWXC1____) {
        r.w[inst->shamt] = rrt;
    } else if (inst->fop == LDXC1____) {
        r.l[inst->shamt / 2] = r64;
    } else if (inst->fop == L_S______) {
        r.w[inst->rt] = rrt;
    } else if (inst->fop == L_D______) {
        r.l[inst->rt / 2] = r64;
    }
}

/**********************************************************************/
void MipsFpu::decode(MipsInst *inst)
{
    switch (inst->opcode) {
    case 0:
        if (inst->funct == 1) {
            inst->fop = (inst->rt & 0x1) ? MOVT_____ : MOVF_____;
            inst->attr = READ_RS | WRITE_RD_COND;
        }
        break;
    case 17:
        if (inst->rs == 8) {
            if (inst->rt & 0x2) {
                inst->fop = (inst->rt & 0x1) ? BC1TL____ : BC1FL____;
                inst->attr = BRANCH_LIKELY;
            } else {
                inst->fop = (inst->rt & 0x1) ? BC1T_____ : BC1F_____;
                inst->attr = BRANCH;
            }
            break;
        }

        switch (inst->funct) {
        case 0:
            switch (inst->rs) {
            case 0:
                inst->fop = MFC1_____;
                inst->attr = READ_NONE | WRITE_RT;
                break;
            case 2:
                inst->fop = CFC1_____;
                inst->attr = READ_NONE | WRITE_RT;
                break;
            case 4:
                inst->fop = MTC1_____;
                inst->attr = READ_RT | WRITE_NONE;
                break;
            case 6:
                inst->fop = CTC1_____;
                inst->attr = READ_RT | WRITE_NONE;
                break;
            case 16:
            case 17:
                inst->fop = (inst->rs == 16) ? ADD_S____ : ADD_D____;
                break;
            }
            break;
        case 1:
            inst->fop = (inst->rs == 16) ? SUB_S____ : SUB_D____;
            break;
        case 2:
            inst->fop = (inst->rs == 16) ? MUL_S____ : MUL_D____;
            break;
        case 3:
            inst->fop = (inst->rs == 16) ? DIV_S____ : DIV_D____;
            break;
        case 4:
            inst->fop = (inst->rs == 16) ? SQRT_S___ : SQRT_D___;
            break;
        case 5:
            inst->fop = (inst->rs == 16) ? ABS_S____ : ABS_D____;
            break;
        case 6:
            inst->fop = (inst->rs == 16) ? MOV_S____ : MOV_D____;
            break;
        case 7:
            inst->fop = (inst->rs == 16) ? NEG_S____ : NEG_D____;
            break;
        case 12:
            inst->fop = (inst->rs == 16) ? ROUND_W_S : ROUND_W_D;
            break;
        case 13:
            inst->fop = (inst->rs == 16) ? TRUNC_W_S : TRUNC_W_D;
            break;
        case 14:
            inst->fop = (inst->rs == 16) ? CEIL_W_S_ : CEIL_W_D_;
            break;
        case 15:
            inst->fop = (inst->rs == 16) ? FLOOR_W_S : FLOOR_W_D;
            break;
        case 17:
            if (inst->rt & 0x1) {
                inst->fop = (inst->rs == 16) ? MOVT_S___ : MOVT_D___;
                break;
            } else {
                inst->fop = (inst->rs == 16) ? MOVF_S___ : MOVF_D___;
                break;
            }
            break;
        case 18:
            inst->fop = (inst->rs == 16) ? MOVZ_S___ : MOVZ_D___;
            inst->attr = READ_RT | WRITE_NONE;
            break;
        case 19:
            inst->fop = (inst->rs == 16) ? MOVN_S___ : MOVN_D___;
            inst->attr = READ_RT | WRITE_NONE;
            break;
        case 21:
            inst->fop = (inst->rs == 16) ? RECIP_S__ : RECIP_D__;
            break;
        case 22:
            inst->fop = (inst->rs == 16) ? RSQRT_S__ : RSQRT_D__;
            break;
        case 32:
            inst->fop = (inst->rs == 17) ? CVT_S_D__ : CVT_S_W__; 
            break;
        case 33:
            inst->fop = (inst->rs == 16) ? CVT_D_S__ : CVT_D_W__; 
            break;
        case 36:
            inst->fop = (inst->rs == 16) ? CVT_W_S__ : CVT_W_D__; 
            break;
        case 48:
            inst->fop = (inst->rs == 16) ? C_F_S____ : C_F_D____;
            break;
        case 49:
            inst->fop = (inst->rs == 16) ? C_UN_S___ : C_UN_D___;
            break;
        case 50:
            inst->fop = (inst->rs == 16) ? C_EQ_S___ : C_EQ_D___;
            break;
        case 51:
            inst->fop = (inst->rs == 16) ? C_UEQ_S__ : C_UEQ_D__;
            break;
        case 52:
            inst->fop = (inst->rs == 16) ? C_OLT_S__ : C_OLT_D__;
            break;
        case 53:
            inst->fop = (inst->rs == 16) ? C_ULT_S__ : C_ULT_D__;
            break;
        case 54:
            inst->fop = (inst->rs == 16) ? C_OLE_S__ : C_OLE_D__;
            break;
        case 55:
            inst->fop = (inst->rs == 16) ? C_ULE_S__ : C_ULE_D__;
            break;
        case 56:
            inst->fop = (inst->rs == 16) ? C_SF_S___ : C_SF_D___;
            break;
        case 58:
            inst->fop = (inst->rs == 16) ? C_SEQ_S__ : C_SEQ_D__;
            break;
        case 59:
            inst->fop = (inst->rs == 16) ? C_NGL_S__ : C_NGL_D__;
            break;
        case 60:
            inst->fop = (inst->rs == 16) ? C_LT_S___ : C_LT_D___;
            break;
        case 61:
            inst->fop = (inst->rs == 16) ? C_NGE_S__ : C_NGE_D__;
            break;
        case 62:
            inst->fop = (inst->rs == 16) ? C_LE_S___ : C_LE_D___;
            break;
        case 63:
            inst->fop = (inst->rs == 16) ? C_NGT_S__ : C_NGT_D__;
            break;
        }
        break;
    case 19:
        switch (inst->funct) {
        case 0:
            inst->fop = LWXC1____;
            inst->attr = READ_RS | READ_RT | LOAD_4B_ALIGN;
            break;
        case 1:
            inst->fop = LDXC1____;
            inst->attr = READ_RS | READ_RT | LOAD_8B;
            break;
        case 8:
            inst->fop = SWXC1____;
            inst->attr = READ_RS | READ_RT | STORE_4B_ALIGN;
            break;
        case 9:
            inst->fop = SDXC1____;
            inst->attr = READ_RS | READ_RT | STORE_8B;
            break;
        case 32:
            inst->fop = MADD_S___;
            break;
        case 33:
            inst->fop = MADD_D___;
            break;
        case 40:
            inst->fop = MSUB_S___;
            break;
        case 41:
            inst->fop = MSUB_D___;
            break;
        case 48:
            inst->fop = NMADD_S__;
            break;
        case 49:
            inst->fop = NMADD_D__;
            break;
        case 56:
            inst->fop = NMSUB_S__;
            break;
        case 57:
            inst->fop = NMSUB_D__;
            break;
        }
        break;
    case 49:
        inst->fop = L_S______;
        inst->attr = READ_RS | LOAD_4B_ALIGN;
        break;
    case 53:
        inst->fop = L_D______;
        inst->attr = READ_RS | LOAD_8B;
        break;
    case 57:
        inst->fop = S_S______;
        inst->attr = READ_RS | STORE_4B_ALIGN;
        break;
    case 61:
        inst->fop = S_D______;
        inst->attr = READ_RS | STORE_8B;
        break;
    }
}


/**********************************************************************/
char* MipsFpu::getinstname(MipsInst *inst)
{
    instname = (char*)setinst(inst->fop);
    return instname;
}

/**********************************************************************/
char* MipsFpu::getmnemonic(MipsInst *inst)
{
    getinstname(inst);
    
    char *arg;
    arg = new char[MNEMONIC_BUF_SIZE];
    arg[0] = '\0';
    int size = MNEMONIC_BUF_SIZE;
    
    switch (inst->fop) {
    case MOVF_____:
    case MOVT_____:
        snprintf(arg, size, "$%d, $%d, $fcc%d",
                 inst->rd, inst->rs, inst->rt >> 2);
        break;
    case MFC1_____:
    case MTC1_____:
        snprintf(arg, size, "$%d, $f%d", inst->rt, inst->rd);
        break;
    case CFC1_____:
    case CTC1_____:
        snprintf(arg, size, "$%d, c%d", inst->rt, inst->rd);
        break;
    case BC1F_____:
    case BC1T_____:
    case BC1FL____:
    case BC1TL____:
        snprintf(arg, size, "$fcc%d, %08x", inst->rt >> 2, 
                 inst->pc + 4 + (exts32(inst->imm, 16) << 2));
        break;
    case ADD_S____:
    case ADD_D____:
    case SUB_S____:
    case SUB_D____:
    case MUL_S____:
    case MUL_D____:
    case DIV_S____:
    case DIV_D____:
        snprintf(arg, size, "$f%d, $f%d, $f%d",
                 inst->shamt, inst->rd, inst->rt);
        break;
    case SQRT_S___:
    case SQRT_D___:
    case ABS_S____:
    case ABS_D____:
    case MOV_S____:
    case MOV_D____:
    case NEG_S____:
    case NEG_D____:
    case ROUND_W_S:
    case ROUND_W_D:
    case TRUNC_W_S:
    case TRUNC_W_D:
    case CEIL_W_S_:
    case CEIL_W_D_:
    case FLOOR_W_S:
    case FLOOR_W_D:
    case RECIP_S__:
    case RECIP_D__:
    case RSQRT_S__:
    case RSQRT_D__:
    case CVT_S_D__:
    case CVT_S_W__:
    case CVT_D_S__:
    case CVT_D_W__:
    case CVT_W_S__:
    case CVT_W_D__:
        snprintf(arg, size, "$f%d, $f%d", inst->shamt, inst->rd);
        break;
    case MOVF_S___:
    case MOVF_D___:
    case MOVT_S___:
    case MOVT_D___:
        snprintf(arg, size, "$f%d, $f%d, $fcc%d",
                 inst->shamt, inst->rd, inst->rt >> 2);
        break;
    case MOVZ_S___:
    case MOVZ_D___:
    case MOVN_S___:
    case MOVN_D___:
        snprintf(arg, size, "$f%d, $f%d, $%d",
                 inst->shamt, inst->rd, inst->rt);
        break;
    case C_F_S____:
    case C_F_D____:
    case C_UN_S___:
    case C_UN_D___:
    case C_EQ_S___:
    case C_EQ_D___:
    case C_UEQ_S__:
    case C_UEQ_D__:
    case C_OLT_S__:
    case C_OLT_D__:
    case C_ULT_S__:
    case C_ULT_D__:
    case C_OLE_S__:
    case C_OLE_D__:
    case C_ULE_S__:
    case C_ULE_D__:
    case C_SF_S___:
    case C_SF_D___:
    case C_SEQ_S__:
    case C_SEQ_D__:
    case C_NGL_S__:
    case C_NGL_D__:
    case C_LT_S___:
    case C_LT_D___:
    case C_NGE_S__:
    case C_NGE_D__:
    case C_LE_S___:
    case C_LE_D___:
    case C_NGT_S__:
    case C_NGT_D__:
        snprintf(arg, size, "$fcc%d, $f%d, $f%d",
                 inst->shamt >> 2, inst->rd, inst->rt);
        break;        
    case LWXC1____:
    case LDXC1____:
        snprintf(arg, size, "$f%d, $%d($%d)", 
                 inst->shamt, inst->rt, inst->rs);
        break;
    case SWXC1____:
    case SDXC1____:
        snprintf(arg, size, "$f%d, $%d($%d)",
                 inst->rd, inst->rt, inst->rs);
        break;
    case MADD_S___:
    case MADD_D___:
    case MSUB_S___:
    case MSUB_D___:
    case NMADD_S__:
    case NMADD_D__:
    case NMSUB_S__:
    case NMSUB_D__:
        snprintf(arg, size, "$f%d, $f%d, $f%d, $f%d",
                 inst->shamt, inst->rs, inst->rd, inst->rt);
        break;
    case L_S______:
    case L_D______:
    case S_S______:
    case S_D______:
        snprintf(arg, size, "$f%d, %d($%d)",
                 inst->rt, exts32(inst->imm, 16), inst->rs);
        break;
    }
    
    snprintf(mnemonic, size, "%-9s%s", instname, arg);
    DELETE_ARRAY(arg);
    return mnemonic;
}

/*********************************************************************/
void MipsFpu::print()
{
    for (int i = 0; i < NFREG; i += 2) {
        printf("              f%.2d=%08x(%12g) "
               "f%.2d=%08x(%12g) (d%.2d=%12g)\n",
               i, r.w[i], r.s[i], i + 1, r.w[i + 1], r.s[i + 1],
               i, r.d[i / 2]);
    }
}

/**********************************************************************/
inline uint032_t exts32(uint032_t x, int y)
{
    if (y == 32)
        return x;
    uint032_t temp = 0xffffffff << y;
    if (x & (1 << (y - 1)))
        return (temp | (x & ~temp));
    else
        return (x & ~temp);
}

/*********************************************************************/
inline const char *setinst(uint fop)
{
    if      (fop == MOVF_____) return "movf";
    else if (fop == MOVT_____) return "movt";
    else if (fop == MFC1_____) return "mfc1";
    else if (fop == CFC1_____) return "cfc1";
    else if (fop == MTC1_____) return "mtc1";
    else if (fop == CTC1_____) return "ctc1";
    else if (fop == BC1F_____) return "bc1f";
    else if (fop == BC1T_____) return "bc1t";
    else if (fop == BC1FL____) return "bc1fl";
    else if (fop == BC1TL____) return "bc1tl";
    else if (fop == ADD_S____) return "add.s";
    else if (fop == ADD_D____) return "add.d";
    else if (fop == SUB_S____) return "sub.s";
    else if (fop == SUB_D____) return "sub.d";
    else if (fop == MUL_S____) return "mul.s";
    else if (fop == MUL_D____) return "mul.d";
    else if (fop == DIV_S____) return "div.s";
    else if (fop == DIV_D____) return "div.d";
    else if (fop == SQRT_S___) return "sqrt.s";
    else if (fop == SQRT_D___) return "sqrt.d";
    else if (fop == ABS_S____) return "abs.s";
    else if (fop == ABS_D____) return "abs.d";
    else if (fop == MOV_S____) return "mov.s";
    else if (fop == MOV_D____) return "mov.d";
    else if (fop == NEG_S____) return "neg.s";
    else if (fop == NEG_D____) return "neg.d";
    else if (fop == ROUND_W_S) return "round.w.s";
    else if (fop == ROUND_W_D) return "round.w.d";
    else if (fop == TRUNC_W_S) return "trunc.w.s";
    else if (fop == TRUNC_W_D) return "trunc.w.d";
    else if (fop == CEIL_W_S_) return "ceil.w.s";
    else if (fop == CEIL_W_D_) return "ceil.w.d";
    else if (fop == FLOOR_W_S) return "floor.w.s";
    else if (fop == FLOOR_W_D) return "floor.w.d";
    else if (fop == MOVF_S___) return "movf.s";
    else if (fop == MOVF_D___) return "movf.d";
    else if (fop == MOVT_S___) return "movt.s";
    else if (fop == MOVT_D___) return "movt.d";
    else if (fop == MOVZ_S___) return "movz.s";
    else if (fop == MOVZ_D___) return "movz.d";
    else if (fop == MOVN_S___) return "movn.s";
    else if (fop == MOVN_D___) return "movn.d";
    else if (fop == RECIP_S__) return "recip.s";
    else if (fop == RECIP_D__) return "recip.d";
    else if (fop == RSQRT_S__) return "rsqrt.s";
    else if (fop == RSQRT_D__) return "rsqrt.d";
    else if (fop == CVT_S_D__) return "cvt.s.d";
    else if (fop == CVT_S_W__) return "cvt.s.w";
    else if (fop == CVT_D_S__) return "cvt.d.s";
    else if (fop == CVT_D_W__) return "cvt.d.w";
    else if (fop == CVT_W_S__) return "cvt.w.s";
    else if (fop == CVT_W_D__) return "cvt.w.d";
    else if (fop == C_F_S____) return "c.f.s";
    else if (fop == C_F_D____) return "c.f.d";
    else if (fop == C_UN_S___) return "c.un.s";
    else if (fop == C_UN_D___) return "c.un.d";
    else if (fop == C_EQ_S___) return "c.eq.s";
    else if (fop == C_EQ_D___) return "c.eq.d";
    else if (fop == C_UEQ_S__) return "c.ueq.s";
    else if (fop == C_UEQ_D__) return "c.ueq.d";
    else if (fop == C_OLT_S__) return "c.olt.s";
    else if (fop == C_OLT_D__) return "c.olt.d";
    else if (fop == C_ULT_S__) return "c.ult.s";
    else if (fop == C_ULT_D__) return "c.ult.d";
    else if (fop == C_OLE_S__) return "c.ole.s";
    else if (fop == C_OLE_D__) return "c.ole.d";
    else if (fop == C_ULE_S__) return "c.ule.s";
    else if (fop == C_ULE_D__) return "c.ule.d";
    else if (fop == C_SF_S___) return "c.sf.s";
    else if (fop == C_SF_D___) return "c.sf.d";
    else if (fop == C_SEQ_S__) return "c.seq.s";
    else if (fop == C_SEQ_D__) return "c.seq.d";
    else if (fop == C_NGL_S__) return "c.ngl.s";
    else if (fop == C_NGL_D__) return "c.ngl.d";
    else if (fop == C_LT_S___) return "c.lt.s";
    else if (fop == C_LT_D___) return "c.lt.d";
    else if (fop == C_NGE_S__) return "c.nge.s";
    else if (fop == C_NGE_D__) return "c.nge.d";
    else if (fop == C_LE_S___) return "c.le.s";
    else if (fop == C_LE_D___) return "c.le.d";
    else if (fop == C_NGT_S__) return "c.ngt.s";
    else if (fop == C_NGT_D__) return "c.ngt.d";
    else if (fop == LWXC1____) return "lwxc1";
    else if (fop == LDXC1____) return "ldxc1";
    else if (fop == SWXC1____) return "swxc1";
    else if (fop == SDXC1____) return "sdxc1";
    else if (fop == MADD_S___) return "madd.s";
    else if (fop == MADD_D___) return "madd.d";
    else if (fop == MSUB_S___) return "msub.s";
    else if (fop == MSUB_D___) return "msub.d";
    else if (fop == NMADD_S__) return "nmadd.s";
    else if (fop == NMADD_D__) return "nmadd.d";
    else if (fop == NMSUB_S__) return "nmsub.s";
    else if (fop == NMSUB_D__) return "nmsub.d";
    else if (fop == L_S______) return "l.s";
    else if (fop == L_D______) return "l.d";
    else if (fop == S_S______) return "s.s";
    else if (fop == S_D______) return "s.d";
    else                       return "";
}

/**********************************************************************/
