#ifndef _LINUX_LIST_H
#define _LINUX_LIST_H

struct list_head {
    struct list_head *next, *prev;
};

#define LIST_HEAD_INIT(name) { &(name), &(name) }

#define LIST_HEAD(name)                           \
    struct list_head name = LIST_HEAD_INIT(name)

static inline void INIT_LIST_HEAD(struct list_head *list)
{
    list->next = list;
    list->prev = list;
}

/**
 * Insert a new entry between two known consecutive entries.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline void __list_add(struct list_head *newptr,
                              struct list_head *prev,
                              struct list_head *next)
{
    next->prev = newptr;
    newptr->next = next;
    newptr->prev = prev;
    prev->next = newptr;
}

/**
 * list_add - add a new entry
 * \param newptr new entry to be added
 * \param head list head to add it after
 *
 * Insert a new entry after the specified head.
 * This is good for implementing stacks.
 */
static inline void list_add(struct list_head *newptr, struct list_head *head)
{
    __list_add(newptr, head, head->next);
}

/**
 * list_add_tail - add a new entry
 * \param newptr new entry to be added
 * \param head list head to add it before
 *
 * Insert a new entry before the specified head.
 * This is useful for implementing queues.
 */
static inline void list_add_tail(struct list_head *newptr, struct list_head *head)
{
    __list_add(newptr, head->prev, head);
}

/**
 * Delete a list entry by making the prev/next entries
 * point to each other.
 *
 * This is only for internal list manipulation where we know
 * the prev/next entries already!
 */
static inline void __list_del(struct list_head * prev, struct list_head * next)
{
    next->prev = prev;
    prev->next = next;
}

/**
 * list_del - deletes entry from list.
 * \param entry the element to delete from the list.
 * Note: list_empty() on entry does not return true after this, the entry is
 * in an undefined state.
 */
static inline void list_del(struct list_head *entry)
{
    __list_del(entry->prev, entry->next);
    entry->next = 0;
    entry->prev = 0;
}

/**
 * list_replace - replace old entry by new one
 * \param old the element to be replaced
 * \param newptr the new element to insert
 *
 * If old was empty, it will be overwritten.
 */
static inline void list_replace(struct list_head *old,
                                struct list_head *newptr)
{
    newptr->next = old->next;
    newptr->next->prev = newptr;
    newptr->prev = old->prev;
    newptr->prev->next = newptr;
}

static inline void list_replace_init(struct list_head *old,
                                     struct list_head *newptr)
{
    list_replace(old, newptr);
    INIT_LIST_HEAD(old);
}

/**
 * list_del_init - deletes entry from list and reinitialize it.
 * \param entry the element to delete from the list.
 */
static inline void list_del_init(struct list_head *entry)
{
    __list_del(entry->prev, entry->next);
    INIT_LIST_HEAD(entry);
}

/**
 * list_move - delete from one list and add as another's head
 * \param list the entry to move
 * \param head the head that will precede our entry
 */
static inline void list_move(struct list_head *list, struct list_head *head)
{
    __list_del(list->prev, list->next);
    list_add(list, head);
}

/**
 * list_move_tail - delete from one list and add as another's tail
 * \param list the entry to move
 * \param head the head that will follow our entry
 */
static inline void list_move_tail(struct list_head *list,
                                  struct list_head *head)
{
    __list_del(list->prev, list->next);
    list_add_tail(list, head);
}

/**
 * list_is_last - tests whether @list is the last entry in list @head
 * \praram list the entry to test
 * \param head the head of the list
 */
static inline int list_is_last(const struct list_head *list,
                               const struct list_head *head)
{
    return list->next == head;
}

/**
 * list_empty - tests whether a list is empty
 * \param head the list to test.
 */
static inline int list_empty(const struct list_head *head)
{
    return head->next == head;
}

/**
 * list_entry - get the struct for this entry
 * \param ptr the &struct list_head pointer.
 * \param type the type of the struct this is embedded in.
 * \param member the name of the list_struct within the struct.
 */
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)

#define container_of(ptr, type, member) ({                            \
            const typeof( ((type *)0)->member ) *__mptr = (ptr);      \
            (type *)( (char *)__mptr - offsetof(type,member) );})

#define list_entry(ptr, type, member)           \
    container_of(ptr, type, member)

/**
 * list_first_entry - get the first element from a list
 * \param ptr the list head to take the element from.
 * \param type the type of the struct this is embedded in.
 * \param member the name of the list_struct within the struct.
 *
 * Note, that list is expected to be not empty.
 */
#define list_first_entry(ptr, type, member)     \
    list_entry((ptr)->next, type, member)

/**
 * list_for_each	-	iterate over a list
 * \param pos the &struct list_head to use as a loop cursor.
 * \param head the head for your list.
 */
#define list_for_each(pos, head)                \
    for (pos = (head)->next; pos != (head);     \
         pos = pos->next)

/**
 * list_for_each_prev	-	iterate over a list backwards
 * @pos:	the &struct list_head to use as a loop cursor.
 * @head:	the head for your list.
 */
#define list_for_each_prev(pos, head)           \
    for (pos = (head)->prev; pos != (head);     \
         pos = pos->prev)

/**
 * list_for_each_entry	-	iterate over list of given type
 * \param pos the type * to use as a loop cursor.
 * \param head the head for your list.
 * \param member the name of the list_struct within the struct.
 */
#define list_for_each_entry(pos, head, member)                        \
    for (pos = list_entry((head)->next, typeof(*pos), member);        \
         &pos->member != (head);                                      \
         pos = list_entry(pos->member.next, typeof(*pos), member))

/**
 * list_for_each_entry_reverse - iterate backwards over list of given type.
 * \param pos the type * to use as a loop cursor.
 * \param head the head for your list.
 * \param member the name of the list_struct within the struct.
 */
#define list_for_each_entry_reverse(pos, head, member)                \
    for (pos = list_entry((head)->prev, typeof(*pos), member);        \
         &pos->member != (head);                                      \
         pos = list_entry(pos->member.prev, typeof(*pos), member))

/**
 * list_prepare_entry - prepare a pos entry for use in list_for_each_entry_continue()
 * \param pos the type * to use as a start point
 * \param head the head of the list
 * \param member the name of the list_struct within the struct.
 *
 * Prepares a pos entry for use as a start point in list_for_each_entry_continue().
 */
#define list_prepare_entry(pos, head, member)          \
    ((pos) ? : list_entry(head, typeof(*pos), member))

#endif
