/**********************************************************************/
/* Many-Core Architecture Research Project       Arch Lab. TOKYO TECH */
/**********************************************************************/
#ifndef _NETWORK_H_
#define _NETWORK_H_

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "core/define.h"
#include "assert.h"
#include "list.h"

#define MC_NAME "SimMc M-Core alpha : Many-Core and NoC Simulator"
#define MC_VER "v1.1 2010-07-30"

#define MC_LIB_VER "1.0.0"

/************************************************************************/
/* MieruPC Memory Map : 0x100 - 0x1ff                    Ver.2009-06-19 */
/************************************************************************/
#define MP_MMAP_BASE                   0x00000100

#define MP_MMAP_KEYCODE                (MP_MMAP_BASE + 0x00)
#define MP_MMAP_LCD                    (MP_MMAP_BASE + 0x04)
#define MP_MMAP_BUZZER                 (MP_MMAP_BASE + 0x08)
#define MP_MMAP_TIMER                  (MP_MMAP_BASE + 0x0c)
#define MP_MMAP_GPIO_0                 (MP_MMAP_BASE + 0xf0)
#define MP_MMAP_GPIO_4                 (MP_MMAP_BASE + 0xf4)
#define MP_MMAP_GPIO_8                 (MP_MMAP_BASE + 0xf8)
#define MP_MMAP_GPIO_C                 (MP_MMAP_BASE + 0xfc)

/************************************************************************/
/* ScalableCore Memory Map : 0x200 - 0x3ff               Ver.2009-06-19 */
/************************************************************************/
#define SC_MMAP_BASE                   0x00000200

#define SC_MMAP_NODE_ID                (SC_MMAP_BASE + 0x00)
#define SC_MMAP_NOC_SIZE               (SC_MMAP_BASE + 0x04)

#define SC_MMAP_PC                     (SC_MMAP_BASE + 0x08)
#define SC_MMAP_TIMER                  (SC_MMAP_BASE + 0x10) // clock
#define SC_MMAP_TIMER_D                (SC_MMAP_BASE + 0x14) // 

#define SC_MMAP_INST_LOOP              (SC_MMAP_BASE + 0x18) // ??? b
#define SC_MMAP_INST_LOOP__            (SC_MMAP_BASE + 0x1c) // ??? nop

#define SC_MMAP_DMA_PUT_REMOTE_ID      (SC_MMAP_BASE + 0x20)
#define SC_MMAP_DMA_PUT_REMOTE_ADDR    (SC_MMAP_BASE + 0x24)
#define SC_MMAP_DMA_PUT_REMOTE_STRIDE  (SC_MMAP_BASE + 0x28)
#define SC_MMAP_DMA_PUT_LOCAL_ADDR     (SC_MMAP_BASE + 0x2c)
#define SC_MMAP_DMA_PUT_LOCAL_STRIDE   (SC_MMAP_BASE + 0x30)
#define SC_MMAP_DMA_PUT_SIZE           (SC_MMAP_BASE + 0x34)
#define SC_MMAP_DMA_PUT_CMD            (SC_MMAP_BASE + 0x3c)

#define SC_MMAP_DMA_GET_REMOTE_ID      (SC_MMAP_BASE + 0x40)
#define SC_MMAP_DMA_GET_REMOTE_ADDR_L  (SC_MMAP_BASE + 0x44)
#define SC_MMAP_DMA_GET_REMOTE_ADDR_H  (SC_MMAP_BASE + 0x48)
#define SC_MMAP_DMA_GET_LOCAL_ADDR     (SC_MMAP_BASE + 0x4c)
#define SC_MMAP_DMA_GET_SIZE           (SC_MMAP_BASE + 0x50)
#define SC_MMAP_DMA_GET_REMOTE_STRIDE  (SC_MMAP_BASE + 0x54)
#define SC_MMAP_DMA_GET_LOCAL_STRIDE   (SC_MMAP_BASE + 0x58)
#define SC_MMAP_DMA_GET_CMD            (SC_MMAP_BASE + 0x5c)

#define SC_MMAP_DMA_FLAG               (SC_MMAP_BASE + 0x80) // for ack/nack

#define SC_MMAP_NODE_SERVER            (SC_MMAP_BASE + 0x100)
#define SC_MMAP_INCC_WRIGHT_ADDR       (SC_MMAP_BASE + 0x104)

/**********************************************************************/
/* gmem memmap */
#define GMEM_BASE                  0x00000200

#define GMEM_DMA_GET_REMOTE_ID     (GMEM_BASE + 0x40)
#define GMEM_DMA_GET_REMOTE_ADDR_L (GMEM_BASE + 0x44)
#define GMEM_DMA_GET_REMOTE_ADDR_H (GMEM_BASE + 0x48)
#define GMEM_DMA_GET_LOCAL_ADDR_L  (GMEM_BASE + 0x4c)
#define GMEM_DMA_GET_LOCAL_ADDR_H  (GMEM_BASE + 0x50)
#define GMEM_DMA_GET_SIZE          (GMEM_BASE + 0x54)
#define GMEM_DMA_GET_REMOTE_STRIDE (GMEM_BASE + 0x58)
#define GMEM_DMA_GET_CMD           (GMEM_BASE + 0x5c)

/********************************************************************/
#define ENV_FLIT_VIEW "FLITVIEW"

#define DEF_RANK_X 4
#define DEF_RANK_Y 4
#define MAX_ROW 127
#define MAX_COL 127
#define MAX_NODE (MAX_ROW * MAX_COL)

#define MAX_LOG_MODE 5
#define DEF_MAX_CYCLE MAX_CYCLE_DEF

#define MAX_FLITS_PER_PACKET 11
#define MAX_MEM_LATENCY (1<<8 - 1)

#define DEFAULT_STRIDE 4

#define IDX(id) ((int)(((id) >> 8) & 0xff))
#define IDY(id) ((int)((id) & 0xff))

#define print_error(...)                        \
    fprintf(stderr, "## ERROR: " __VA_ARGS__)
#define print_exit(...)                         \
    (print_error(__VA_ARGS__), exit(1))

#define round_robin(__i,last, MAX)                                      \
    for(int __i=((last)+1>=(MAX))?0:(last)+1, __t=0;                    \
        __t<(MAX);                                                      \
        __i=(__i+1>=(MAX))?0:__i+1, __t++)

#define opposite(x)                             \
    ((x) == N ? S :                             \
     ((x) == E ? W :                            \
      ((x) == S ? N :                           \
       ((x) == W ? E :                          \
        0xffff))))

#define LOG1_HEADER                                                     \
    "## cycle dma_put        src-id dst-id   src-addr   dst-addr      byte stride\n" \
    "## cycle dfma_put_4b    src-id dst-id       data   dst-addr\n"    \
    "## cycle recieve-header            id                                 stride\n" \
    "## cycle recieve-addr              id              dst-addr\n"     \
    "## cycle receive-body              id       data\n"

typedef unsigned short uint016_t;

/**
 * flags of flit
 */
enum {
    FLIT_VALID = 1 << 5, /** valid */
    FLIT_HEADER = 1 << 4, /** header flit */
    FLIT_ADDR = 1 << 3, /** address flit */
    FLIT_STRIDE = 1 << 2, /** stride flit */
    FLIT_DATA = 1 << 1, /** data flit */
    FLIT_TAILER = 1 << 0 /** tailer flit */
};

/**
 * 
 */
enum {
    N, /** valid */
    E,
    S,
    W,
    C,
    NUM_DIRECTION,
};

enum {
    DMAFLAG_NA,
    DMAFLAG_ACK,
    DMAFLAG_NACK,
};

enum {
    NODE_COMP,
    NODE_SERVER,
    NODE_MEM,
    NODE_PATH,
};

typedef struct {
    uint032_t clk, phase;
    uint032_t y, x;
    uint032_t direction;
    uint032_t site;
    uint032_t data;
} log_t;

typedef struct {
    uint032_t clk;
    uint008_t type;
    uint008_t y;
    uint008_t x;
    uint008_t dir_and_site;
    uint032_t data;
} log5_t;

typedef union {
    uint008_t type;
    struct {
        uint008_t type;
        uint016_t coreid;
        uint016_t tag;
    } header;
    struct {
        uint008_t type; 
        uint032_t data;
    } body;
} flit_t;

typedef enum {
    MC_CHAR,
    MC_INT,
    MC_FLOAT,
    MC_DOUBLE,
} MC_Datatype;

typedef enum {
    MC_MAX,
    MC_MIN,
    MC_SUM,
    MC_PROD,
    MC_LAND,
    MC_BAND,
    MC_LOR,
    MC_BOR,
    MC_LXOR,
    MC_BXOR,
    MC_MAXLOC,
    MC_MINLOC,
} MC_Op;

const flit_t flit_invalid = {0};

/********************************************************************/
typedef Board Core;
class Router;
class Incc;
class Incc32;
class Incc64;
class Node;
class Nodeinfo;
class GMemory;
class Cmdbuf;
class Bitmap;
class Getdata;

class Nodeinfo {
public:
    int id;
    int type;
    int state;
    char *filename;
    char *option;
    Nodeinfo();
};

class Env {
private:
    uint phase;
    FILE *fp;
    int checkarg(int, char **);
    ullint gettime();
    void print_log4();
    void print_log5();
    int check_mclib(int, const char *g);
    
public:
    Core *core;
    Router *router;
    Incc32 *incc32;
    Incc64 *incc64;
    Node *node;
    GMemory *gmem;

    Nodeinfo *nodeinfo;
    uint032_t *prev_pc;

    int add_module;

    int incc_activity;
    FILE *plot;

    int noc_col, noc_row;
    int rank_x, rank_y, nocrank;
    int nr_node;
    int nr_node_comp;
    uint debug_mode, log_mode, quiet_mode, server_mode;
    ullint maxcycle;
    ullint cycle;
    char *logfile;
    int multicycle;
    char path_prefix[128];

    Env();
    ~Env();
	void make_nodeinfo(char *filename);
    int initialize(int, char **);
    int initialize(void);
    void execute();
    void finalize();
    void print_log();
    uint032_t entry2id(int);
    int id2entry(uint032_t);
    int id2entry(int, int);
    int id2id(int, int);
    const char *int2type(int type);
    void print_flit(flit_t*);

    int finishcond();
	void core_update();
	void memory_update();
	void incc_update();
	void router_update();
	void router_switching();
	void comm_update();
	void post_update();


};

class Inbuf {
private:
    int _head, _tail, _num_entry;
public:
    static const int BUF_SIZE = 4;
    flit_t buf[BUF_SIZE];

    uint008_t request;
    Inbuf();
    bool isempty();
    bool isfull();
    int num_entry();
    bool deq();
    bool enq(flit_t);
    flit_t *head();
    void rtcomp(uint032_t id);
};

typedef struct {
    uint032_t remote_id;
    uint032_t remote_addr;
    uint032_t remote_stride;
    uint032_t local_addr;
    uint032_t local_stride;
    uint032_t size;
    uint032_t reserved;
    uint032_t cmd;
} commcmd_t;

class Cmdbuf {
    typedef commcmd_t elem_t;
private:
    elem_t *buf;
    int _head, _tail, _num_entry, nentry;
    
public:
    Cmdbuf(int);
    ~Cmdbuf();
    bool isempty();
    bool isfull();
    int num_entry();
    bool deq();
    bool enq(void*);
    void *head();
};

class Incc {
protected:
    Env *env;
    unsigned long long clock_begin;

public:
    int *coreid;

    uint008_t irdy, irdy_;
    uint008_t ordy, ordy_;

    commcmd_t cmd_send;
    commcmd_t cmd_recv;

    /* dynamic router */
    flit_t in;
    flit_t inbuf, inbuf_;
    flit_t outbuf, _outbuf;
    flit_t out;

    int dma_type, dma_type_, _dma_type;
    int dma_send_state, _dma_send_state;
    int dma_send_state_;
    uint032_t send_count;

    Router *router;

    Incc();
    virtual ~Incc();
    virtual void init(Env*, int) = 0;
    virtual void update() = 0;
    virtual void comm() = 0;
    virtual void print_packet() = 0;
    virtual void print_flit() = 0;
};

class Incc32 : public Incc {
    // wires
    int nackentry;
    
    commcmd_t _getreqbuf;
    Cmdbuf *getreqbuf;
    Bitmap *nackmap;

    // getreqbuf size
    static const unsigned int GETREQBUF_SIZE = 2;

public:
    // dma send state
    enum {
        INIT,
        ACK_FH, ACK_FA, ACK_FD, 
        NACK_FH, NACK_FA, NACK_FD, 
        PUT_FH, PUT_FA, PUT_FS, PUT_FD,
        GET_FH, GET_FA, GET_FS, GET_FD,
        PUT64_FH, PUT64_FAL, PUT64_FAH, PUT64_FD,
        NUM_STATE,
    };
    
    // type of dma
    enum {
        DMA_NA,
        DMA_PUT,
        DMA_GET,
        DMA_ACK,
        DMA_NACK,
        DMA_NUM_CMD,
    };
    
    commcmd_t *core_incc_if;

    MainMemory *mem;
    void plot();
    void print_MClib();
    const char *int2dmatype(int);
    const char *int2state(int);
    void print_flit();
    void print_packet();

    uint032_t *memmap;
    void init(Env*, int);
    void update();
    void comm();

    Incc32();
    ~Incc32();
};

class Getdata {
    static const unsigned int BUF_SIZE = 1024;
    static const unsigned int NENTRY = 128;

    typedef struct {
        uint032_t read_idx;
        uint064_t read_addr;
        uint032_t size;
        uint032_t data[BUF_SIZE];
    } getdata_t;

    getdata_t getdata[NENTRY];

    GMemory *mem;
    
public:

    int _head, _tail, _num_entry, nentry;

    Getdata(GMemory*);
    bool enq(uint064_t address, uint032_t size);
    bool deq();
    bool isfull();
    bool isempty();
    uint032_t nextdata();
};

class Incc64 : public Incc {
    // wires
    int nackentry;

    commcmd_t *core_incc_if;
    
    commcmd_t _getreqbuf;
    Cmdbuf *getreqbuf;

    GMemory *mem;
    uint032_t *memmap;

    void print_flit();
    void print_packet();
    
    Getdata *getdata;

    int readable;

    unsigned int ready_counter;

public:
    // dma send state
    enum {
        INIT,
        GET_FH, GET_FA, GET_FS, GET_FD,
        ACK_FH, ACK_FA, ACK_FD,
        NACK_FH, NACK_FA, NACK_FD,
        NUM_STATE,
    };

    // type of dma
    enum {
        DMA_NA,
        DMA_null,
        DMA_GET,
        DMA_ACK,
        DMA_NACK,
        DMA_NUM_CMD,
    };

    // getreqbuf size
    static const unsigned int GETREQBUF_SIZE = 2;

    void init(Env*, int);
    void update();
    void comm();

    Bitmap *nackmap;

    const char *int2dmatype(int);
    const char *int2state(int);

    Incc64();
    ~Incc64();
};

class Router {
private:
    Env *env;
    uint008_t last_send[NUM_DIRECTION];
    uint008_t do_round_robin[NUM_DIRECTION];

public:
    int id;
    Router *router[NUM_DIRECTION - 1];
    Incc *incc;
    Inbuf in_buf[NUM_DIRECTION];
    uint008_t irdy[NUM_DIRECTION];
    uint008_t ordy[NUM_DIRECTION];
    uint008_t ordy_[NUM_DIRECTION];
    flit_t out[NUM_DIRECTION];
    flit_t in[NUM_DIRECTION];
    
    Router();
    ~Router();
    void init(Env*, int);
    void update();
    void switching();
    void comm();
    void print(FILE*);
};

class Node {
public:
    Core *core;
    void *mem;
    Router *router;
    Incc *incc;
    Node();
};

class GMemory {
public:
    static const unsigned int GMEM_PAGE_BITS = 12;
    static const uint064_t GMEM_PAGE_SIZE = 1llu << GMEM_PAGE_BITS;
    static const uint064_t GMEM_PAGE_MASK = GMEM_PAGE_SIZE - 1;
    static const unsigned int PAGE_LATENCY = 80; // cycle
    static const unsigned int READ_LATENCY = 20; // cycle
    
private:
    typedef struct {
        struct list_head list;
        uint064_t pagehead;
        uint032_t data[GMEM_PAGE_SIZE/sizeof(uint032_t)];
    } page_t;
    
    page_t *newpage(const uint064_t);
    uint032_t *setpageentry(const uint064_t);
    void delpage(const uint064_t);

    uint032_t *pagebuf;
    unsigned int latency_counter;

public:
    GMemory();
    ~GMemory();
    struct list_head alist;
    uint032_t *getrealaddr(const uint064_t);
    void read4b(const uint064_t, uint032_t*);
    void write4b(const uint064_t, const uint032_t);
    uint032_t *getpagerealaddr(const uint064_t addr);
    int pagehit(const uint064_t addr);
};

class Bitmap {
public:
    static const unsigned int NOT_FOUND = -1u;

    unsigned int *bitmap;
    unsigned int nbits;
    int words;

    Bitmap(unsigned int); // allocate bitmap
    ~Bitmap();
    int isclear();
    void set(unsigned int); // set a bit (ex. 001000 => 3)
    void reset(unsigned int); // reset a bit (ex. 000100 => 2)
    int lsb(); // search lsb 
    int msb(); // search msb
    static int lsb(unsigned int*, unsigned int); // search lsb
    static int msb(unsigned int*, unsigned int); // search msb
    static int lsb(unsigned int); // search lsb
    static int msb(unsigned int); // search msb
    void print(); // search msb
};

#endif
